/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/* @(#)66       1.4  src/avs/fs/mmfs/ts/kernext/gpl-linux/ktrccalls.c, mmfs, avs_rgpfs32, rgpfs32s006a 5/12/08 15:34:50 */

/**************************************************************************
 *
 * Common kernel trace functions
 *
 **************************************************************************/

#ifndef GPFS_PRINTF

#ifndef __KERNEL__
#  define __KERNEL__
#endif

#ifdef GPFS_LINUX

#define Kprint printk 
#include <Shark-gpl.h>

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <stdarg.h>
#include <linux/string.h>

#include <Trace.h>
#include <lxtrace.h>
#include <verdep.h>

#if LINUX_KERNEL_VERSION > 2060900
 #include <linux/hardirq.h> /* interrupt_context */
#else
 #include <asm/hardirq.h> /* interrupt_context */
#endif

unsigned int interrupt_context()
{
  return in_interrupt();
}

#elif defined(GPFS_SOLARIS)

#include <Trace.h>
#include <lxtrace.h>
#include <sys/cmn_err.h>
#include <sys/cpuvar.h>
#include <string.h>
#include <stdarg.h>

#define Kprint(...) cmn_err (CE_NOTE,__VA_ARGS__);
#define MIN(a,b)  ((a)<(b) ? (a) : (b))

unsigned int interrupt_context()
{
  return CPU_ON_INTR(CPU);
}

typedef unsigned int Boolean;
#define false 0
#define true 1

#endif /* GPFS_SOLARIS */ 


#if defined(__64BIT__) || defined(_LP64)
# define Int64 long long
# define ARGLEN 8
# define ARGTYPE Int64
#else
# define Int32 int
# define ARGLEN 4
# define ARGTYPE Int32
#endif

char stringPadding[8];

extern trcdev_header_t   lxthe;
extern void trc_append_record(const void*, size_t);
extern Boolean isTraced(uint);
extern int trc_start_record(trc_datahdr_t *, size_t, int);
extern void trc_end_record();

static void
_STraceArgs(int* trRecLenP, int* stringLenP, int nArgs, int pos, va_list listP)
{
  int dataLen;
  int i;
  ARGTYPE tmpint;
  char *s;
  int stringLen;
  int stringPadLen;

  dataLen = 0;

  /* Handle argument lists that include a string parameter */
  if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
  {
    /* Items (if any) preceeding the string argument */
    for (i = 0; i < pos; i++)
    {
      tmpint = va_arg(listP, ARGTYPE);
      trc_append_record(&tmpint, ARGLEN);
      dataLen += ARGLEN;
    }

    /* Copy the string, making sure it does not overflow the buffer */
    s = va_arg(listP, char*);
    if (s == NULL)
      s = "(NULL)";
    if (s < (char*)4096) /* bad address */
    {
      Kprint("_STrace: bad address 0x%X hook 0x%X\n", s, lxthe.hdrP->trHook);
      s = "<bad address>";
    }
    stringLen = strlen(s);
    stringLen = MIN(stringLen,
                    LXTRACE_MAX_DATA - sizeof(trc_datahdr_t) -
                      (nArgs*ARGLEN) - 1 - (ARGLEN-1));
    trc_append_record(s, stringLen);
    stringPadLen = ARGLEN - (stringLen%ARGLEN);
    trc_append_record(stringPadding, stringPadLen);
    *stringLenP = stringLen + stringPadLen;
    dataLen += stringLen + stringPadLen;

    /* Append items following string argument */
    for (i = pos; i < nArgs; i++)
    {
      tmpint = va_arg(listP, ARGTYPE);
      trc_append_record(&tmpint, ARGLEN);
      dataLen += ARGLEN;
    }
  }
  else /* !IS_SFORMAT */
  {
    /* Place the fixed parameters in the temporary trace buffer */
    for (i = 0; i < nArgs; i++)
    {
      tmpint = va_arg(listP, ARGTYPE);
      trc_append_record(&tmpint, ARGLEN);
      dataLen += ARGLEN;
    }
    *stringLenP = 0;
  }

  /* Append the float argument */
  if (pos == _TR_FORMAT_F)
  {
    /* Although the argument is really a double, don't tell the compiler,
       so that it will not generate code using floating point hardware
       that is not supposed to be used in the kernel. */
    /* double tmpdbl = va_arg(listP, double); */
    unsigned long long tmpdbl = va_arg(listP, unsigned long long);
    trc_append_record(&tmpdbl, sizeof(tmpdbl));
    dataLen += sizeof(tmpdbl);
  }

  *trRecLenP = sizeof(trc_datahdr_t) + dataLen;
  /* DBGASSERT(*trRecLenP <= LXTRACE_MAX_DATA); */
}


void _STraceNB(int hookword, int nArgs, int pos, ...)
{
  trc_datahdr_t hdr;
  int recLen;
  int rc;
  va_list listP;
  int trRecLen;
  int stringLen;

  /* Trace calls from interrupt level are not supported.  If anybody needs
     them, changing writeLock to use spin_lock_irqsave should be all that
     is needed to allow non-blocking traces to work. */
  if (interrupt_context())
    return;

  if (!isTraced(hookword))
    return;

  /* Test for trace formats that aren't supported yet */
  if ((pos == _TR_FORMAT_I) && (nArgs > LXTRACE_MAX_FORMAT_SUBS))
  {
#ifdef DBGASSERTS
    Kprint("_STrace: too many arguments (hook %X)\n", hookword);
#endif /* DBGASSERTS */
    return;
  }

  /* Build a data header and append it to the trace file.  If there is a
     string, the length is not yet known, so use the maximum.  It will be
     patched to the correct value later. */
  hdr.trHook = hookword;
  hdr.trNArgs = nArgs;
  hdr.trSPos = pos;
  hdr.trSLen = 0;   /* invalid if there is a string; fix below */
  if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
    recLen = LXTRACE_MAX_DATA;
  else
  {
    recLen = sizeof(hdr) + nArgs*ARGLEN;
    if (pos == _TR_FORMAT_F)
      recLen += ARGLEN;
  }
  rc = trc_start_record(&hdr, recLen, true);

  /* If the header was successfully written, collect arguments directly into
     the trace buffer */
  if (rc == 0)
  {
    va_start(listP, pos);
    _STraceArgs(&trRecLen, &stringLen, nArgs, pos, listP);
    va_end(listP);

    /* Patch the string and record lengths now that the string has been
       copied */
    lxthe.hdrP->trSLen = stringLen;
    lxthe.tHdrP->trLength = trRecLen;

    /* Trace record complete */
    trc_end_record();
  }
}

void _STrace(int hookword, int nArgs, int pos, ...)
{
  trc_datahdr_t hdr;
  int recLen;
  int rc;
  va_list listP;
  int trRecLen;
  int stringLen;

  /* Trace calls from interrupt level are not supported.  If anybody needs
     them, changing writeLock to use spin_lock_irqsave should be all that
     is needed to allow non-blocking traces to work. */
  if (interrupt_context())
    return;

  if (!isTraced(hookword))
    return;

  /* Test for trace formats that aren't supported yet */
  if ((pos == _TR_FORMAT_I) && (nArgs > LXTRACE_MAX_FORMAT_SUBS))
  {
#ifdef DBGASSERTS
    Kprint("_STrace: too many arguments (hook %X)\n", hookword);
#endif /* DBGASSERTS */
    return;
  }

  /* Build a data header and append it to the trace file.  If there is a
     string, the length is not yet known, so use the maximum.  It will be
     patched to the correct value later. */
  hdr.trHook = hookword;
  hdr.trNArgs = nArgs;
  hdr.trSPos = pos;
  hdr.trSLen = 0;   /* invalid if there is a string; fix below */
  if (pos >= 0  &&  pos < LXTRACE_MAX_FORMAT_SUBS)
    recLen = LXTRACE_MAX_DATA;
  else
  {
    recLen = sizeof(hdr) + nArgs*ARGLEN;
    if (pos == _TR_FORMAT_F)
      recLen += ARGLEN;
  }
  rc = trc_start_record(&hdr, recLen, false);

  /* If the header was successfully written, collect arguments directly into
     the trace buffer */
  if (rc == 0)
  {
    va_start(listP, pos);
    _STraceArgs(&trRecLen, &stringLen, nArgs, pos, listP);
    va_end(listP);

    /* Patch the string and record lengths now that the string has been
       copied */
    lxthe.hdrP->trSLen = stringLen;
    lxthe.tHdrP->trLength = trRecLen;

    /* Trace record complete */
    trc_end_record();
  }
}

void _XTraceNB(int hookword, char *fmt, ...)
{
  trc_datahdr_t hdr;
  int rc;
  va_list vargs;
  int stringLen;

  /* Trace calls from interrupt level are not supported.  If anybody needs
     them, changing writeLock to use spin_lock_irqsave should be all that
     is needed to allow non-blocking traces to work. */
  if (interrupt_context())
    return;

  if (!isTraced(hookword))
    return;

  /* Build a data header and append it to the trace file.  Since the length
     is not yet known, use the maximum.  It will be patched to the correct
     value later. */
  hdr.trHook = hookword;
  hdr.trNArgs = 0;
  hdr.trSPos = _TR_FORMAT_X;
  hdr.trSLen = -1;   /* invalid; fix below */
  rc = trc_start_record(&hdr, LXTRACE_MAX_DATA, true);

  /* If the header was successfully written, format the string directly
     into the trace buffer */
  if (rc == 0)
  {
    va_start(vargs, fmt);
    stringLen = vsnprintf(lxthe.writeBuf.nextP,
                          LXTRACE_MAX_DATA-sizeof(trc_datahdr_t), fmt, vargs) + 1;
    va_end(vargs);
    if (stringLen > LXTRACE_MAX_DATA-sizeof(trc_datahdr_t))
    {
      Kprint("_XTraceNB: argument too long.  len=%d max=%d hook=0x%X\n",
             stringLen, LXTRACE_MAX_DATA-sizeof(trc_datahdr_t)-1, hookword);
      stringLen = LXTRACE_MAX_DATA-sizeof(trc_datahdr_t);
    }

    /* Patch the string and record lengths now that vsnprintf has calculated
       the length that it formatted */
    lxthe.hdrP->trSLen = ((stringLen+ARGLEN-1)/ARGLEN)*ARGLEN;
    lxthe.tHdrP->trLength = sizeof(hdr) + lxthe.hdrP->trSLen;

    /* Advance pointer into trace buffer by the length of the string just
       appended */
    lxthe.writeBuf.nextP += lxthe.hdrP->trSLen;

    /* Trace record complete */
    trc_end_record();
  }
}

void _XTrace(int hookword, char *fmt, ...)
{
  trc_datahdr_t hdr;
  int rc;
  va_list vargs;
  int stringLen;

  /* Trace calls from interrupt level are not supported.  If anybody needs
     them, changing writeLock to use spin_lock_irqsave should be all that
     is needed to allow non-blocking traces to work. */
  if (interrupt_context())
    return;

  if (!isTraced(hookword))
    return;

  /* Build a data header and append it to the trace file.  Since the length
     is not yet known, use the maximum.  It will be patched to the correct
     value later. */
  hdr.trHook = hookword;
  hdr.trNArgs = 0;
  hdr.trSPos = _TR_FORMAT_X;
  hdr.trSLen = -1;   /* invalid; fix below */
  rc = trc_start_record(&hdr, LXTRACE_MAX_DATA, false);

  /* If the header was successfully written, format the string directly
     into the trace buffer */
  if (rc == 0)
  {
    va_start(vargs, fmt);
    stringLen = vsnprintf(lxthe.writeBuf.nextP,
                          LXTRACE_MAX_DATA-sizeof(trc_datahdr_t), fmt, vargs) + 1;
    va_end(vargs);
    if (stringLen > LXTRACE_MAX_DATA-sizeof(trc_datahdr_t))
    {
      Kprint("_XTrace: argument too long.  len=%d max=%d hook=0x%X\n",
             stringLen, LXTRACE_MAX_DATA-sizeof(trc_datahdr_t)-1, hookword);
      stringLen = LXTRACE_MAX_DATA-sizeof(trc_datahdr_t);
    }

    /* Patch the string and record lengths now that vsnprintf has calculated
       the length that it formatted */
    lxthe.hdrP->trSLen = ((stringLen+ARGLEN-1)/ARGLEN)*ARGLEN;
    lxthe.tHdrP->trLength = sizeof(hdr) + lxthe.hdrP->trSLen;

    /* Advance pointer into trace buffer by the length of the string just
       appended */
    lxthe.writeBuf.nextP += lxthe.hdrP->trSLen;

    /* Trace record complete */
    trc_end_record();
  }
}
#endif /* GPFS_PRINTF */
